#include "Ball.h"
#include <list>

const float damp = 0.80f;
const float topSpin = 0.25f;
const float maxSpeed = 0.25f;

CBall::CBall()
{
	radius = 1.2f;
	sphere = 0;
	world = 0;
	displayList = 0;
}

CBall::~CBall()
{
	if (sphere)
		gluDeleteQuadric(sphere);
	//baseTexture.Delete();
	reflectTexture.Delete();
	particleTexture.Delete();

	if (displayList != 0)
		glDeleteLists(displayList, 1);
}

void CBall::Init()
{
	if (!sphere)
	{
		sphere = gluNewQuadric();
		gluQuadricTexture(sphere, GL_TRUE);
		gluQuadricNormals(sphere, GLU_SMOOTH);
	}

	//baseTexture.LoadBitmapFile("blueplasma.bmp");
	reflectTexture.LoadBitmapFile("light.bmp");
	particleTexture.LoadBitmapFile("particle.bmp");
	baseTexture = &reflectTexture;

	Reset();

	aura.SetNumberOfParticles(20);
	aura.SetEmissionRate(15);
	aura.SetColor(0.7,0.7,0.7,0.0,0.0,1.0);
	aura.SetLife(2.25f);
	aura.SetSize(1.75f,1.25f);
	aura.SetFade(1.75f);
	aura.SetFadeRange(0.25f);
	aura.SetOriginRange(CVector3d(0.05,0.05,0.05));
	aura.SetVelocity(CVector3d(0,2.0,0));
	aura.SetVelocityRange(CVector3d(2,2,2));
	aura.SetTexture(&particleTexture);

	displayList = glGenLists(1);
	glNewList(displayList, GL_COMPILE);
	glEnable(GL_TEXTURE_2D);
	glDisable(GL_LIGHT1);
	glColor4f( 1.0, 1.0, 1.0, 1.0f );		
	glDepthMask(GL_FALSE);
	gluSphere(sphere, radius, 32, 32 );
	glDepthMask(GL_TRUE);
	
	reflectTexture.Bind();
	glEnable(GL_BLEND);
	glColor4f( 1.0, 1.0, 1.0, 0.65 );	
	glBlendFunc(GL_SRC_ALPHA, GL_ONE);
	glEnable(GL_TEXTURE_GEN_S); 
	glEnable(GL_TEXTURE_GEN_T);
	gluSphere(sphere, radius, 32, 32 );
	glDisable(GL_TEXTURE_GEN_S); 
	glDisable(GL_TEXTURE_GEN_T);
	glDisable(GL_BLEND);

	glEndList();
}

void CBall::Reset()
{
	if (world)
		position = world->startPosition + CVector3d(0,radius,0);

	groundHeight = position.y();
	safePosition = position;
	done = false;
	force.Zero();
	velocity.Zero();
	externalForce.Zero();
	aura.Flush();

	// Set up identity matrix
	rotation[0] = 1.0f;
	rotation[1] = 0.0f;
	rotation[2] = 0.0f;
	rotation[3] = 0.0f;
	rotation[4] = 0.0f;
	rotation[5] = 1.0f;
	rotation[6] = 0.0f;
	rotation[7] = 0.0f;
	rotation[8] = 0.0f;
	rotation[9] = 0.0f;
	rotation[10] = 1.0f;
	rotation[11] = 0.0f;
	rotation[12] = 0.0f;
	rotation[13] = 0.0f;
	rotation[14] = 0.0f;
	rotation[15] = 1.0f;
}

const CVector3d& CBall::GetPosition() 
{
	return position;
}

const CVector3d& CBall::GetSafePosition() 
{
	return safePosition;
}

float CBall::GetRadius()
{
	return radius;
}

float CBall::GetGroundHeight()
{
	return groundHeight;
}

void CBall::SetWorld(CWorld* world)
{
	this->world = world;
}

void CBall::SetDone(bool d)
{
	done = d;
}

bool CBall::IsDone()
{
	return done;
}

void CBall::Update(float elapsed)
{
	// Apply gravity
	//CVector3d gforce = CVector3d(0.0f, -0.00025f, 0.0f);
  CVector3d gforce = CVector3d(0.0f, -0.003f, 0.0f);
	velocity += gforce;
	velocity += force;
	velocity += externalForce;
	float speed = velocity.Length();

	if (speed > maxSpeed)
		velocity *= maxSpeed/speed;

	//velocity *= 0.9985f;
	velocity *= 0.9965f;
	externalForce *= 0.95f;

	if (done)
	{
		velocity.SetX(velocity.x()*0.985);
		velocity.SetZ(velocity.z()*0.985);
	}

	position += velocity;

	CheckCollisions();

	CalculateRotation();	

	// Check if they fell off the edge
	if (position.y() <= groundHeight - 30.0f)
	{
		velocity.SetY(0.0f);
		velocity.Normalize();
		position = safePosition - velocity*radius*0.5f;
		force.Zero();
		velocity.Zero();
	}
}

void CBall::SetBall(const CTexture *tex, float r, float g, float b)
{
	baseTexture = tex;
	aura.SetColor(0.7,0.7,0.7,r,g,b);
	aura.Flush();
}

void CBall::SetForce(const CVector3d& f)
{
	// Dont let them control it when its falling
	//if (fabs(velocity.y()) < 0.1f)
	  //force = f * 0.0002f;
	  force = f * 0.0008f;
}

void CBall::ApplyForce(const CVector3d& f)
{
	externalForce += f * 0.0008f;
}

void CBall::Draw()
{
	glPushAttrib(GL_ENABLE_BIT);
	glPushMatrix();
	
	glTranslatef( position.x(), position.y(), position.z() );

	if (done)
	{
		glDepthMask(0);  
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA,GL_ONE);
		glEnable(GL_TEXTURE_2D);
		aura.Draw(0.02f,true);
		glDisable(GL_BLEND);
		glDepthMask(1);
	}

	glMultMatrixd( rotation );
	baseTexture->Bind();
	glCallList(displayList);

	glPopMatrix();
	glPopAttrib();
}

void CBall::CalculateRotation()
{
	// Calculate the axis to rotate around
	CVector3d arbitraryAxis( velocity.z(), 0, -velocity.x() );
	float length = arbitraryAxis.Length();
	if (length != 0.0f)
	{
		double angle = 60.0f * (length/radius);
		arbitraryAxis /= length;	
		// Use opengl to calculate the balls rotation
		glMatrixMode( GL_MODELVIEW );
		glPushMatrix();
		 glLoadIdentity();
		 glRotatef( angle, arbitraryAxis.x(), arbitraryAxis.y(), arbitraryAxis.z() );
		 glMultMatrixd( rotation );
		 glGetDoublev(GL_MODELVIEW_MATRIX, rotation );
		glPopMatrix();
	}
}

void CBall::CheckCollisions()
{
	if (world == 0)
		return;
	
	std::list<CBoundingVolume*>::iterator bound;

	for (bound = world->bounds.begin(); bound != world->bounds.end(); bound++)
	{
		CVector3d intersection, normal;
		if ((*bound)->CheckSphereCollision(position, radius, intersection, normal))
		{
			velocity.Reflect(normal);
			position = intersection+normal*radius;			

			// Check if we are on flat ground
			if (normal.y() == 1.0f)
				safePosition = position;					
			if (normal.y() > 0.0f)
				groundHeight = intersection.y()+radius;
			// Damp bouncing if they fall from high
			if (velocity.y() > 0.02f)
				velocity.SetY(velocity.y()*damp);
			if (normal.y() == 0.0f)
				velocity *= 0.9;
		}	
	}	
}
