#include "BoundingBox.h"

CBoundingBox::CBoundingBox(const CVector3d& c, const CVector3d& s)
:	CBoundingVolume(c)
{
	size = s;
	rightAxis = CVector3d(1.0f,0.0f,0.0f);
	upAxis = CVector3d(0.0f,1.0f,0.0f);
	normal = CVector3d(0.0f,0.0f,1.0f);

	pts[0] = c + CVector3d( -size.x(),  size.y(),  size.z() );
	pts[1] = c + CVector3d( -size.x(), -size.y(),  size.z() );
	pts[2] = c + CVector3d(  size.x(), -size.y(),  size.z() );
	pts[3] = c + CVector3d(  size.x(),  size.y(),  size.z() );
	pts[4] = c + CVector3d( -size.x(),  size.y(), -size.z() );
	pts[5] = c + CVector3d( -size.x(), -size.y(), -size.z() );
	pts[6] = c + CVector3d(  size.x(), -size.y(), -size.z() );
	pts[7] = c + CVector3d(  size.x(),  size.y(), -size.z() );
}

CBoundingBox::CBoundingBox(const CVector3d& c, const CVector3d& s, const CVector3d& right, const CVector3d& up)
:	CBoundingVolume(c)
{
	size = s;
	rightAxis = right.UnitLength();
	upAxis = up.UnitLength();
	normal = CVector3d::CrossProduct(upAxis, rightAxis);

	pts[0] = c - rightAxis*size.x() + upAxis*size.y() + normal*size.z();
	pts[1] = c - rightAxis*size.x() - upAxis*size.y() + normal*size.z();
	pts[2] = c + rightAxis*size.x() - upAxis*size.y() + normal*size.z();
	pts[3] = c + rightAxis*size.x() + upAxis*size.y() + normal*size.z();
	pts[4] = c - rightAxis*size.x() + upAxis*size.y() - normal*size.z();
	pts[5] = c - rightAxis*size.x() - upAxis*size.y() - normal*size.z();
	pts[6] = c + rightAxis*size.x() - upAxis*size.y() - normal*size.z();
	pts[7] = c + rightAxis*size.x() + upAxis*size.y() - normal*size.z();
}

const CVector3d& CBoundingBox::GetSize() const
{
	return size;
}

const CVector3d& CBoundingBox::GetNormal() const
{
	return normal;
}

const CVector3d& CBoundingBox::GetRightAxis() const
{
	return rightAxis;
}

const CVector3d& CBoundingBox::GetUpAxis() const
{
	return upAxis;
}

void CBoundingBox::SetCenter(const CVector3d& c)
{
	CVector3d diff = c-center;
 	center = c;
	for (int i = 0; i < 8; i++)
		pts[i] += diff;
}

bool CBoundingBox::InFrustum(const CFrustum& frustum) const
{
	return frustum.CheckBox(pts);
}

bool CBoundingBox::CheckSphereCollision(const CVector3d& spherePosition, float radius, CVector3d& intersection, CVector3d& collisionNormal) const
{
	const float* boxSize = (const float*)size;

	float alignedDist[3];
	CVector3d distance = spherePosition - center;
	alignedDist[0] = distance*rightAxis;
	alignedDist[1] = distance*upAxis;
	alignedDist[2] = distance*normal;

	float totalDistSqrd = 0.0f;
	for( int i = 0; i < 3; i++ ) 
	{
		if( alignedDist[i] < -boxSize[i] ) 
		{
			float dist = alignedDist[i] + boxSize[i];
			totalDistSqrd += dist*dist;  
		}
		else if( alignedDist[i] > boxSize[i] ) 
		{
			float dist = alignedDist[i] - boxSize[i];
			totalDistSqrd += dist*dist;  
		}
	}

	// If we have collided, then figure out where
	if( totalDistSqrd <= radius*radius ) 
	{
		float sign[3] = { 1.0f,1.0f,1.0f };
		
		for( int i = 0; i < 3; i++ ) 
		{
			if ( alignedDist[i] < 0.0f )
			{
				alignedDist[i] = -alignedDist[i];
				sign[i] = -1.0f;
			}
		}
		
		if ( alignedDist[0] <= boxSize[0] )
		{
			if ( alignedDist[1] <= boxSize[1] )
			{
				if ( alignedDist[2] <= boxSize[2])
				{
					// inside - shouldn't ever happen
					// but does sometimes with the moving floor
					// just treat it as a positive y face
					intersection = (center + rightAxis*alignedDist[0]*sign[0] + upAxis*boxSize[1]*sign[1] + normal*alignedDist[2]*sign[2]);				
					collisionNormal = upAxis;	
				}
				else
				{
					// Z face
					intersection = (center + rightAxis*alignedDist[0]*sign[0] + upAxis*alignedDist[1]*sign[1] + normal*boxSize[2]*sign[2]);				
					collisionNormal = (spherePosition - intersection).UnitLength();	
				}
			}
			else
			{
				if ( alignedDist[2] <= boxSize[2])
				{
					// Y face
					intersection = (center + rightAxis*alignedDist[0]*sign[0] + upAxis*boxSize[1]*sign[1] + normal*alignedDist[2]*sign[2]);				
					collisionNormal = (spherePosition - intersection).UnitLength();	
				}
				else
				{
					// YZ edge
					intersection = (center+ rightAxis*alignedDist[0]*sign[0] + upAxis*boxSize[1]*sign[1] + normal*boxSize[2]*sign[2]);				
					collisionNormal = (spherePosition - intersection).UnitLength();
				}
			}
		}
		else
		{
			if ( alignedDist[1] <= boxSize[1] )
			{
				if ( alignedDist[2] <= boxSize[2])
				{
					// X face
					intersection = (center + rightAxis*boxSize[0]*sign[0] + upAxis*alignedDist[1]*sign[1] + normal*alignedDist[2]*sign[2]);				
					collisionNormal = (spherePosition - intersection).UnitLength();
				}
				else
				{
					// XZ edge
					intersection = (center+ rightAxis*boxSize[0]*sign[0] + upAxis*alignedDist[1]*sign[1] + normal*boxSize[2]*sign[2]);
					collisionNormal = (spherePosition - intersection).UnitLength();
				}
			}
			else
			{
				if ( alignedDist[2] <= boxSize[2])
				{
					// XY edge
					intersection = (center+ rightAxis*boxSize[0]*sign[0] + upAxis*boxSize[1]*sign[1] + normal*alignedDist[2]*sign[2]);
					collisionNormal = (spherePosition - intersection).UnitLength();
				}
				else
				{
					// XYZ Vertex
					intersection = (center+ rightAxis*boxSize[0]*sign[0] + upAxis*boxSize[1]*sign[1] + normal*boxSize[2]*sign[2]);
					collisionNormal = (spherePosition - intersection).UnitLength();	
				}
			}
		}
		return true;
	}
	else
		return false;	
}

