#include "BoundingPlane.h"

CBoundingPlane::CBoundingPlane(const CVector3d& p1, const CVector3d& p2, const CVector3d& p3)
{
	center = (p3+p1)/2.0f;
	rightAxis = (p3-p2);
	upAxis = (p1-p2);

	pts[0] = p1;
	pts[1] = p2;
	pts[2] = p3;
	pts[3] = p1+rightAxis;

	float rightLength = rightAxis.Length();
	float upLength = upAxis.Length();
	size.Set(rightLength/2.0f, upLength/2.0f, 0.0f);

	// normalize the axes
	if (rightLength != 0.0f)
		rightAxis /= rightLength;
	if (upLength != 0.0f)
		upAxis /= upLength;
  
	normal = CVector3d::CrossProduct(rightAxis, upAxis);
}

const CVector3d& CBoundingPlane::GetSize() const
{
	return size;
}

const CVector3d& CBoundingPlane::GetNormal() const
{
	return normal;
}

const CVector3d& CBoundingPlane::GetRightAxis() const
{
	return rightAxis;
}

const CVector3d& CBoundingPlane::GetUpAxis() const
{
	return upAxis;
}

void CBoundingPlane::SetCenter(const CVector3d& c)
{
	CVector3d diff = c - center;
 	center = c;
	for (int i = 0; i < 4; i++)
		pts[i] += diff;
}

bool CBoundingPlane::InFrustum(const CFrustum& frustum) const
{
	return frustum.CheckPlane(pts);
}

bool CBoundingPlane::CheckSphereCollision(const CVector3d& spherePosition, float radius, CVector3d& intersection, CVector3d& collisionNormal) const
{
	const float* planeSize = (const float*)size;

	float alignedDist[3];
	CVector3d distance = spherePosition - center;
	alignedDist[0] = distance*rightAxis;
	alignedDist[1] = distance*upAxis;
	alignedDist[2] = distance*normal;

	float totalDistSqrd = 0.0f;
	for( int i = 0; i < 3; i++ ) 
	{
		if( alignedDist[i] < -planeSize[i] ) 
		{
			float dist = alignedDist[i] + planeSize[i];
			totalDistSqrd += dist*dist;  
		}
		else if( alignedDist[i] > planeSize[i] ) 
		{
			float dist = alignedDist[i] - planeSize[i];
			totalDistSqrd += dist*dist;  
		}
	}

	// If we have collided, then respond
	if( totalDistSqrd <= radius*radius ) 
	{
		float sign[2] = { 1.0f,1.0f };
		
		for( int i = 0; i < 2; i++ ) 
		{
			if ( alignedDist[i] < 0.0f )
			{
				alignedDist[i] = -alignedDist[i];
				sign[i] = -1.0f;
			}
		}
		
		if ( alignedDist[0] <= planeSize[0] )
		{
			if ( alignedDist[1] <= planeSize[1] )
			{
				// Z face
				intersection = (center+ rightAxis*alignedDist[0]*sign[0] + upAxis*alignedDist[1]*sign[1]);				
				collisionNormal = (spherePosition - intersection).UnitLength();	
			}
			else
			{
				// YZ edge
				intersection = (center+ rightAxis*alignedDist[0]*sign[0] + upAxis*planeSize[1]*sign[1]);				
				collisionNormal = (spherePosition - intersection).UnitLength();
			}
		}
		else
		{
			if ( alignedDist[1] <= planeSize[1] )
			{
				// XZ edge
				intersection = (center+ rightAxis*planeSize[0]*sign[0] + upAxis*alignedDist[1]*sign[1]);
				collisionNormal = (spherePosition - intersection).UnitLength();
			}
			else
			{
				// XYZ Vertex
				intersection = (center+ rightAxis*planeSize[0]*sign[0] + upAxis*planeSize[1]*sign[1]);
				collisionNormal = (spherePosition - intersection).UnitLength();	
			}
		}
		return true;
	}
	else
		return false;
}

