#include "ParticleSystem.h"

CParticleSystem::CParticleSystem()
{
	particles = 0;
	numberOfParticles = 0;
	maxNumberParticles = 0;
	elapsedTime = 0.0f;

	life = 1.0f; 
	fadeRange = 0.01f;	
	fade = 0.05f;
	rate = 0.0f;

	origin.Zero();
	originRange.Zero();
	velocity.Zero();
	velocityRange.Zero();
	force.Zero();

	startColor[0] = startColor[1] = startColor[2] = 1.0f;
	endColor[0] = endColor[1] = endColor[2] = 0.0f;
	startSize = endSize = 1.0f;

	texture = 0;
}

CParticleSystem::~CParticleSystem()
{
	CleanUp();
}

void CParticleSystem::CleanUp()
{
	if (particles != 0)
	{
		delete[] particles;
		particles = 0;
	}
}

void CParticleSystem::SetNumberOfParticles(int num)
{
	maxNumberParticles = num;
	CleanUp();
	particles = new CParticle[maxNumberParticles];
}

void CParticleSystem::SetColor(float start[], float end[])
{
	startColor[0] = start[0];
	startColor[1] = start[1];
	startColor[2] = start[2];
	endColor[0] = end[0];
	endColor[1] = end[1];
	endColor[2] = end[2];
}

void CParticleSystem::SetColor(float r1,float g1, float b1, float r2, float g2, float b2)
{
	startColor[0] = r1;
	startColor[1] = g1;
	startColor[2] = b1;
	endColor[0] = r2;
	endColor[1] = g2;
	endColor[2] = b2;
}

void CParticleSystem::Flush()
{
	numberOfParticles = 0;
	for (int i = 0; i < maxNumberParticles; i++) 
		particles[i].life = 0.0f;
}

void CParticleSystem::Init(bool update)
{	
	elapsedTime = 0.0f;
	numberOfParticles = 0;
	for (int i = 0; i < maxNumberParticles; i++) // Initialize all the particles attributes
	{
		if (rate == 0.0f) // If Rate is zero, then we start all particles at once
		{
			InitParticle(i);
			if (update) // Update each particle a bunch of times if we don't want a burst
				for (int n = 0; n < 25; n++)
					UpdateParticle(i,1.0f);
		}
		else  // Otherwise make all particles dead for now
			particles[i].life = 0.0f;
	}
}

void CParticleSystem::InitParticle(int i)
{
	numberOfParticles++;

	particles[i].life = life; 
	particles[i].fade = fade + frand()*fadeRange; 
	if (particles[i].fade <= 0.0f)
		particles[i].fade = -particles[i].fade;

	particles[i].position = origin + CVector3d(frand()*originRange.x(),frand()*originRange.y(),frand()*originRange.z());

	particles[i].velocity = velocity + CVector3d(frand()*velocityRange.x(),frand()*velocityRange.y(),frand()*velocityRange.z());

	particles[i].r = startColor[0];
	particles[i].g = startColor[1];
	particles[i].b = startColor[2];

	particles[i].size = startSize;
	
	float intervals= particles[i].fade/life*2;
	particles[i].dr = (endColor[0]-startColor[0])*intervals;   
	particles[i].dg = (endColor[1]-startColor[1])*intervals;
	particles[i].db = (endColor[2]-startColor[2])*intervals;
	particles[i].dsize = (endSize-startSize)*intervals;
}

void CParticleSystem::DecrementLife(int i, float elapsed)
{
	particles[i].life -= particles[i].fade*elapsed;
}

void CParticleSystem::AdvanceParticleInList(int i)
{
	particles[i] = particles[numberOfParticles];
}

void CParticleSystem::UpdateAttributes(int i, float elapsed)
{
	particles[i].position += particles[i].velocity*elapsed;
	particles[i].velocity += force*elapsed;
	
	particles[i].r += particles[i].dr*elapsed; 
	particles[i].g += particles[i].dg*elapsed;
	particles[i].b += particles[i].db*elapsed;
	particles[i].size += particles[i].dsize*elapsed;
}

void CParticleSystem::UpdateParticle(int i,float elapsed)
{
	DecrementLife(i,elapsed);
	if (particles[i].life <= 0)    
	{															
		--numberOfParticles;
		if (rate == 0.0f)	// If we are not emmitting at a set rate
			InitParticle(i);	// We can automatically start a new particle
		else if (numberOfParticles > 0) // Otherwise, if there are still particles 
			AdvanceParticleInList(i);     // move the last one into the dead ones spot
	}
	else                      // If not dead
		UpdateAttributes(i,elapsed); // Update it
}

void CParticleSystem::Draw(float elapsed, bool update)
{
	elapsedTime += elapsed;
	if (rate != 0.0f) 
	{
		while (elapsedTime >= rate && numberOfParticles < maxNumberParticles-1)
		{				
			elapsedTime -= rate; 
			InitParticle(numberOfParticles); 
		}
	}

	texture->Bind();

	float modelview[16];
	glGetFloatv(GL_MODELVIEW_MATRIX , modelview);
	// Get our right and up vectors that allow us to align textures to camera
	right = CVector3d(modelview[0],modelview[4],modelview[8]);
	up = CVector3d(modelview[1],modelview[5],modelview[9]);
	
	for (int i = 0; i < numberOfParticles; i++) 
	{
		if (particles[i].life > 0.0f) // Only draw if alive
			DrawParticle(i);
		if (update) // Update particle only if specified
			UpdateParticle(i,elapsed);
	}
}

void CParticleSystem::DrawParticle(int i)
{
	glColor4f(particles[i].r,particles[i].g,particles[i].b,particles[i].life); 
	CParticle *p = &(particles[i]);
	glBegin(GL_TRIANGLE_STRIP);

	 CVector3d point(p->position + (right + up) * p->size);
	 glTexCoord2d(1, 1); glVertex3fv(point);

	 point = p->position + (up - right) * p->size;
   glTexCoord2f(0, 1); glVertex3fv(point);

	 point = p->position + (right - up) * p->size;
	 glTexCoord2f(1, 0); glVertex3fv(point);

	 point = p->position + (right + up) * -(p->size);
	 glTexCoord2f(0, 0); glVertex3fv(point);
	glEnd();
}

