#ifndef H_VECTOR3D_H
#define H_VECTOR3D_H

#include <math.h>

class CVector3d
{
public:
	// Constructors
	CVector3d() { Set(0.0, 0.0, 0.0); }
	CVector3d(float x, float y, float z) { Set(x,y,z); }
	CVector3d(const CVector3d& other) { Set(other); }

	// Set the vector to given 3-tuple
	void Set(float x, float y, float z) { vec[0] = x; vec[1] = y; vec[2] = z; };
	void Set(const CVector3d& other) { vec[0] = other.vec[0]; vec[1] = other.vec[1]; vec[2] = other.vec[2]; }

	// Define a cast operator that returns a pointer to the vec array so that 
	// we can make geometry calls using glVertex3fv, glNormal3fv, etc..
	operator const float *() { return vec; }
  operator const float *() const { return vec; }

	float x() const { return vec[0]; }
  float y() const { return vec[1]; }
  float z() const { return vec[2]; }
	void SetX(float x) { vec[0] = x; };
	void SetY(float y) { vec[1] = y; };
	void SetZ(float z) { vec[2] = z; };
	void Zero() { vec[0] = 0.0f; vec[1] = 0.0f; vec[2] = 0.0f; }; 
	bool IsZero() const { return (vec[0] == 0.0f && vec[1] == 0.0f && vec[2] == 0.0f ); }
	
	float Length() const
	{
		return sqrt(vec[0]*vec[0]+vec[1]*vec[1]+vec[2]*vec[2]);
	}

	float LengthSquared() const
	{
		return (vec[0]*vec[0]+vec[1]*vec[1]+vec[2]*vec[2]);
	}

	CVector3d UnitLength() const
	{
		float length = Length();
		if (length == 0.0)
			return CVector3d(*this);
		else
			return CVector3d( vec[0]/length, vec[1]/length, vec[2]/length );
	}

	const CVector3d& Normalize()
	{
		float length = Length();
		if (length != 0.0)
		{
			vec[0] /= length; 
			vec[1] /= length;
			vec[2] /= length;
		}
		return *this;
	}

	// Cross Product
	CVector3d CrossProduct(const CVector3d& v2) const
	{
		return CVector3d( vec[1]*v2.vec[2] - vec[2]*v2.vec[1], 
			vec[2]*v2.vec[0] - vec[0]*v2.vec[2], 
			vec[0]*v2.vec[1] - vec[1]*v2.vec[0]);
	}

	static CVector3d CrossProduct(const CVector3d& v1, const CVector3d& v2)
	{
		return CVector3d( v1.vec[1]*v2.vec[2] - v1.vec[2]*v2.vec[1],
			v1.vec[2]*v2.vec[0] - v1.vec[0]*v2.vec[2],
			v1.vec[0]*v2.vec[1] - v1.vec[1]*v2.vec[0]);
	}

	// Dot Product
	float operator*(const CVector3d& v2) const
	{
		return ( vec[0]*v2.vec[0] + vec[1]*v2.vec[1] + vec[2]*v2.vec[2] );
	}

	static float DotProduct(const CVector3d& v1, const CVector3d& v2)
	{
		return ( v1.vec[0]*v2.vec[0] - v1.vec[1]*v2.vec[1] + v1.vec[2]*v2.vec[2]);
	}

	// Reflect over a plane's normal
	const CVector3d& Reflect(const CVector3d& normal)
	{
		float speed = Length();
		if (speed != 0.0f)
		{
			CVector3d vector(*this);
			vector /= speed;
			vector = normal*2.0f*(-vector * normal) + vector; 
			vector *= speed;													
			*this = vector;			
		}
		return *this;
	}

	// Assignment operator
	const CVector3d& operator=(const CVector3d& other)
	{
		Set(other);
		return *this;
	}

	// Vector Equality
	bool operator==(const CVector3d& other) const
	{
		return (vec[0] == other.vec[0] && vec[1] == other.vec[1] && vec[2] == other.vec[2]);
	}

	// Vector Addition
	CVector3d operator+(const CVector3d& other) const
	{
		return CVector3d(vec[0] + other.vec[0], vec[1] + other.vec[1], vec[2] + other.vec[2]);
	}

	const CVector3d& operator+=(const CVector3d& other)
	{
		vec[0] += other.vec[0];
		vec[1] += other.vec[1];
		vec[2] += other.vec[2];
		return *this;
	}

	// Vector subtraction
	CVector3d operator-(const CVector3d& other) const
	{
		return CVector3d(vec[0] - other.vec[0], vec[1] - other.vec[1], vec[2] - other.vec[2]);
	}

	const CVector3d& operator-=(const CVector3d& other)
	{
		vec[0] -= other.vec[0];
		vec[1] -= other.vec[1];
		vec[2] -= other.vec[2];
		return *this;
	}

	// Negation
	CVector3d operator-() const
	{
		return CVector3d(-vec[0], -vec[1], -vec[2]);
	}

	// Scalar Multiply
	CVector3d operator*(float scale) const
	{
		return CVector3d( scale*vec[0], scale*vec[1], scale*vec[2]);
	}

	const CVector3d& operator*=(float scale)
	{
		vec[0] *= scale;
		vec[1] *= scale;
		vec[2] *= scale;
		return *this;
	}

	// Scalar Divide
	CVector3d operator/(float scale) const
	{
		if (scale != 0.0)
			return CVector3d( vec[0]/scale, vec[1]/scale, vec[2]/scale);
		else
			return *this;
	}

	const CVector3d& operator/=(float scale)
	{
		if (scale != 0.0)
		{
			vec[0] /= scale;
			vec[1] /= scale;
			vec[2] /= scale;
		}
		return *this;
	}

private:
	float vec[3];
};

#endif
