#include <GL/glut.h>
#include <list>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "FpsCounter.h"
#include "Camera.h"
#include "Ball.h"
#include "Mouse.h"
#include "Texture.h"
#include "Vector3d.h"
#include "World.h"
#include "SkyBox.h"
#include "Clock.h"
#include "BallSelector.h"

CCamera* Camera;
CBall* Ball;
CWorld* World;
CMouse* Mouse;
CSkyBox* SkyBox;
CBallSelector* BallSelector;
CFpsCounter FpsCounter;
CClock Clock;

// Window width and height to calculate aspect ratio
int windowWidth = 800;
int windowHeight = 600;

bool showFPS = false;
bool fullScreen = false;
bool skybox = true;
bool selection = true;

void DisplayCallback()
{
  glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();
	Camera->LookAt(selection);
	
	// World light, doesn't move
	float light0_pos[] = {40.0f, 40.0f, 0.0, 1.0};
	glLightfv( GL_LIGHT0, GL_POSITION, light0_pos );

	if (selection)
	{
		BallSelector->Draw();
	}
	else
	{
		if (skybox)
		{
			glPushMatrix();
			CVector3d offset = Ball->GetPosition();
			glTranslatef(offset.x(),0,offset.z());
			SkyBox->Draw();
			glPopMatrix();
		}
			
		// Draw the floor
		World->DrawFloor();

		// Draw real objects
		// Flip the drawing order so that particle systems show correctly at the end of the level
		if (Ball->IsDone())
		{
			World->DrawObjects();
			Ball->Draw();	
		}
		else
		{
			Ball->Draw();	
			World->DrawObjects();
		}
			
		Clock.Display(windowWidth, windowHeight);
	}

	if (showFPS)
		FpsCounter.Display(windowWidth, windowHeight);

  // draw the buffer to the screen
  glutSwapBuffers();
}

void ResizeCallback(int w, int h)
{
  glViewport( 0, 0, w, h );
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();

	if ( h == 0 )
		h = 1;

	gluPerspective( 35, (float)w/(float)h , 1.0, 500.0 );

	windowWidth = w;
	windowHeight = h;
}

void Init()
{
	// Light0 is a world light that stays put
	GLfloat light0_amb[] = { 0.2, 0.2, 0.2, 1.0 };
  GLfloat light0_dif[] = { 0.5, 0.5, 0.5, 1.0 };
	GLfloat light0_spec[] = { 0.8, 0.8, 0.8, 1.0 };
	glLightfv( GL_LIGHT0, GL_AMBIENT, light0_amb );
  glLightfv( GL_LIGHT0, GL_DIFFUSE, light0_dif );
	glLightfv( GL_LIGHT0, GL_SPECULAR, light0_spec );
	glEnable( GL_LIGHT0 );

	// Light1 is a spotlight that moves with the camera
	GLfloat light1_amb[] = { 0.2, 0.2, 0.2, 1.0 };
	GLfloat light1_dif[] = { 0.5, 0.5, 0.5, 1.0 };
	GLfloat light1_spec[] = { 0.8, 0.8, 0.8, 1.0 };
  GLfloat light1_pos[] = { 0.3, 0.3, 0.3, 1.0 };  
  glLightfv( GL_LIGHT1, GL_POSITION, light1_pos );
  glLightfv( GL_LIGHT1, GL_AMBIENT, light1_amb );
  glLightfv( GL_LIGHT1, GL_DIFFUSE, light1_dif );
  glLightfv( GL_LIGHT1, GL_SPECULAR, light1_spec );
	glLighti( GL_LIGHT1, GL_SPOT_CUTOFF, 60 );
	glLighti( GL_LIGHT1, GL_SPOT_EXPONENT, 10 );
  glEnable( GL_LIGHT1 );

	glEnable( GL_LIGHTING );
  glEnable( GL_COLOR_MATERIAL );
	glMateriali(GL_FRONT, GL_SHININESS, 20);	
	glColorMaterial( GL_FRONT, GL_AMBIENT_AND_DIFFUSE );

	glEnable( GL_DEPTH_TEST );
	glEnable( GL_CULL_FACE );
	glCullFace(GL_BACK);
	glClearStencil(0);
	glClearColor(0,0,0,0);
	glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
	glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
	glHint(GL_POLYGON_SMOOTH_HINT, GL_NICEST);

	Ball = new CBall();
	World = new CWorld();
	Camera = new CCamera();
	Mouse = new CMouse();
	SkyBox = new CSkyBox();
	BallSelector = new CBallSelector();

	Ball->SetWorld(World);
	World->SetBall(Ball);
	Camera->SetTarget(Ball);
	World->Init();
	Ball->Init();
	SkyBox->Init();	
	BallSelector->Init();
	BallSelector->SetBall(Ball);
}

void Reset()
{
	if (selection)
		BallSelector->Reset();
	Ball->Reset();
	Clock.Reset();
	Camera->Reset();	
}

void CleanUp()
{
	delete Ball;
	delete World;
	delete Camera;
	delete Mouse;
	delete SkyBox;
	delete BallSelector;
}

void MouseCallback(int button, int state, int x, int y)
{
	if (selection)
	{
		if (button == GLUT_LEFT_BUTTON && state == GLUT_DOWN)
			BallSelector->ProcessClick(x, windowHeight-y);
	}
	else
	{
		if (button == GLUT_LEFT_BUTTON)
		{
			if (state == GLUT_DOWN)
				Mouse->LeftButtonDown(x, y);
			else
				Mouse->LeftButtonUp();
		}
		else if (button == GLUT_RIGHT_BUTTON)
		{
			if (state == GLUT_DOWN)
				Mouse->RightButtonDown(x, y);
			else
			{
				Mouse->RightButtonUp();
				Ball->SetForce(CVector3d(0.0f, 0.0f, 0.0f));
			}
		}
	}
}

void MouseMotionCallback(int x, int y)
{
	Mouse->TrackMotion(x, y);
}

void IdleCallback()
{
	int time = FpsCounter.GetElapsedTime();
	//printf("%d\n", time);

	if (!selection)
	{
		if (!Ball->IsDone())
		{
			// Handle camera and ball movement here
			if (Mouse->IsLeftButtonDown())
			{
				int deltaX = Mouse->DeltaX(false);
				int deltaY = Mouse->DeltaY(false);
				if (Mouse->IsRightButtonDown())
					Camera->Zoom(-deltaY);
				else
				{
					Camera->Tilt(-deltaY);
					Camera->Rotate(deltaX);
				}		
			}
			else if (Mouse->IsRightButtonDown())
			{
				int deltaX = Mouse->DeltaX();
				int deltaY = -Mouse->DeltaY();
				float theta = Camera->GetRotationDegrees();
				float x = cos(theta)*deltaX + sin(theta)*deltaY;
				float y = -sin(theta)*deltaX + cos(theta)*deltaY;
				Ball->SetForce(CVector3d( x, 0, y ));

				if (!Clock.IsRunning())
					Clock.Start();
			}

			if (World->IsBallDone())
			{
				Ball->SetDone(true);
				Clock.Stop();
			}
		}
		else
		{
			Camera->Rotate(12);
			Camera->Tilt(4);
			Camera->Zoom(4);
		}

		Ball->Update(1.0f);
	}
	else
	{
		Camera->SetTilt(60);
		Camera->Rotate(15);
		if (BallSelector->IsDone())
		{
			selection = false;
			Camera->Reset();
		}
	}

	glutPostRedisplay();
}

void KeyboardCallback(unsigned char c, int, int) 
{
  switch (c) 
	{
  case 27 : // Exit on Escape key
		CleanUp();
		if (fullScreen)
			glutLeaveGameMode();
    exit (0);
    break;
	case 'r' : // Reset to start of level
		Reset();
		break;
	case 'e' : // Reset to ball selection
		selection = true;
		Reset();
		break;
	case 'f' : // Toggle fps counter
		showFPS = !showFPS;
		break;
	/*case 's' : // Toggle reflection
		skybox = !skybox;
		break;*/
	}

	if (Ball->IsDone())
	{
		selection = true;
		Reset();
	}

  glutPostRedisplay();
}

int main(int argc, char **argv)
{
  // create window and rendering context
  glutInit( &argc, argv );
	glutInitDisplayMode( GLUT_DEPTH | GLUT_RGB | GLUT_DOUBLE | GLUT_STENCIL );

   if (!fullScreen)
   {
     glutInitWindowSize( windowWidth, windowHeight );
     glutCreateWindow( "Marble Mayhem" );
   }

   else
   {
     glutGameModeString("800x600:32");
     // enter full screen
     if (glutGameModeGet(GLUT_GAME_MODE_POSSIBLE)) 
       glutEnterGameMode();
     else 
     {
       printf("The game mode is not available\n");
       exit(1);	
     }
   }

  // register display callback
  glutDisplayFunc( DisplayCallback );
  glutKeyboardFunc( KeyboardCallback );
	glutReshapeFunc( ResizeCallback );
	glutIdleFunc( IdleCallback );
	glutMouseFunc( MouseCallback );
	glutMotionFunc( MouseMotionCallback );
 
	Init();

  // pass control over to GLUT
  glutMainLoop();
  
  return 0;       // never reached
}

